import polars as pl

from wimsey import dataframe


def test_that_describe_returns_expected_dictionary_for_df() -> None:
    df = pl.DataFrame({"a": [1.2, 1.3, 1.4], "b": ["one", "two", None]})
    actual = dataframe.describe(df)
    assert 1.29 < actual["mean_a"] < 1.31
    assert actual["null_count_b"] == 1
    assert 0.332 < actual["null_percentage_b"] < 0.334
    assert actual["length"] == 3
    assert actual["columns"] == "a_^&^_b"


def test_that_describe_excludes_non_specified_columns() -> None:
    df = pl.DataFrame({"a": [1.2, 1.3, 1.4], "b": ["one", "two", None]})
    actual = dataframe.describe(df, columns=["a"])
    assert "mean_a" in actual
    assert "mean_b" not in actual


def test_that_describe_excludes_non_specified_metrics() -> None:
    df = pl.DataFrame({"a": [1.2, 1.3, 1.4], "b": ["one", "two", None]})
    actual = dataframe.describe(df, metrics=["mean", "max"])
    assert "mean_a" in actual
    assert "mean_b" in actual
    assert "max_a" in actual
    assert "max_b" in actual
    assert "std_a" not in actual
    assert "std_b" not in actual


def test_that_describe_excludes_non_specified_column_and_metric_combos() -> None:
    df = pl.DataFrame({"a": [1.2, 1.3, 1.4], "b": ["one", "two", None]})
    actual = dataframe.describe(df, columns=["a"], metrics=["count"])
    assert "count_a" in actual
    assert "count_b" not in actual
    assert "min_a" not in actual



from typing import Callable

from wimsey import tests


def test_that_all_possible_tests_are_functions_that_return_partials() -> None:
    for test_name, actual_test in tests.possible_tests.items():
        assert isinstance(
            actual_test(column="anything", other_column="anythin_else"), Callable
        )


def test_all_possible_tests_exposed_as_variables_of_the_same_name_in_module() -> None:
    for test_name in tests.possible_tests:
        assert isinstance(getattr(tests, test_name), Callable)


def test_average_ratio_to_other_column_should_matches_expected() -> None:
    test = tests.average_ratio_to_other_column_should(
        "a", "b", be_greater_than=0.09, be_less_than=0.11
    )
    passing_result = test({"mean_a": 13, "mean_b": 130})
    failing_result = test({"mean_a": 13, "mean_b": 160})
    assert passing_result.success
    assert not failing_result.success



from functools import partial
from typing import Any, Callable
from dataclasses import dataclass


@dataclass
class result:
    name: str
    success: bool
    unexpected: Any = None


def _range_check(metric: str) -> Callable:
    """
    Factory function for generated tests of the form "x should be within range"

    Tests are also factories in themselves, they'll generate functions to take
    only a "describe" object.
    """

    def should(
        describe: dict[str, Any],
        column: str,
        be_exactly: float | int | None = None,
        be_less_than: float | int | None = None,
        be_less_than_or_equal_to: float | int | None = None,
        be_greater_than: float | int | None = None,
        be_greater_than_or_equal_to: float | int | None = None,
    ) -> result:
        """Test that column metric is within designated range"""
        checks: list[bool] = []
        value: Any = describe[f"{metric}_{column}"]
        if be_exactly is not None:
            checks.append(value == be_exactly)
        if be_less_than is not None:
            checks.append(value < be_less_than)
        if be_greater_than is not None:
            checks.append(value > be_greater_than)
        if be_less_than_or_equal_to is not None:
            checks.append(value <= be_less_than_or_equal_to)
        if be_greater_than_or_equal_to is not None:
            checks.append(value >= be_greater_than_or_equal_to)
        return result(
            name=f"{metric}-of-{column}",
            success=all(checks),
            unexpected=value if not all(checks) else None,
        )

    def should_be_partial(
        column: str,
        be_exactly: float | int | None = None,
        be_less_than: float | int | None = None,
        be_less_than_or_equal_to: float | int | None = None,
        be_greater_than: float | int | None = None,
        be_greater_than_or_equal_to: float | int | None = None,
        **kwargs,
    ) -> Callable:
        """Test that column {metric} is within designated range"""
        return_partial = partial(
            should,
            column=column,
            be_exactly=be_exactly,
            be_less_than=be_less_than,
            be_less_than_or_equal_to=be_less_than_or_equal_to,
            be_greater_than=be_greater_than,
            be_greater_than_or_equal_to=be_greater_than_or_equal_to,
        )
        return_partial.required_metrics = {metric}
        return return_partial

    should_be_partial.__doc__ = should_be_partial.__doc__.replace("{metric}", metric)
    return should_be_partial


def row_count_should(
    be_less_than: float | int | None = None,
    be_less_than_or_equal_to: float | int | None = None,
    be_greater_than: float | int | None = None,
    be_greater_than_or_equal_to: float | int | None = None,
    be_exactly: float | int | None = None,
    **kwargs,
) -> Callable:
    """Test that dataframe row count is within designated range"""

    def row_count_should_be(
        description: dict,
        be_less_than: float | int | None = None,
        be_less_than_or_equal_to: float | int | None = None,
        be_greater_than: float | int | None = None,
        be_greater_than_or_equal_to: float | int | None = None,
        be_exactly: float | int | None = None,
    ) -> result:
        """Test that dataframe row count is within designated range"""
        checks: list[bool] = []
        value: float | int = description["length"]
        if be_exactly is not None:
            checks.append(value == be_exactly)
        if be_less_than is not None:
            checks.append(value < be_less_than)
        if be_greater_than is not None:
            checks.append(value > be_greater_than)
        if be_less_than_or_equal_to is not None:
            checks.append(value <= be_less_than_or_equal_to)
        if be_greater_than_or_equal_to is not None:
            checks.append(value >= be_greater_than_or_equal_to)
        if be_exactly is not None:
            checks.append(value == be_exactly)
        return result(
            name="row-count",
            success=all(checks),
            unexpected=value if not all(checks) else None,
        )

    should_be_partial = partial(
        row_count_should_be,
        be_less_than=be_less_than,
        be_less_than_or_equal_to=be_less_than_or_equal_to,
        be_greater_than=be_greater_than,
        be_greater_than_or_equal_to=be_greater_than_or_equal_to,
    )
    should_be_partial.required_metrics = {"length"}
    return should_be_partial


def columns_should(
    have: list[str] | str | None = None,
    not_have: list[str] | str | None = None,
    be: list[str] | str | None = None,
    **kwargs,
) -> Callable:
    """Test column names match up with expected values"""

    def should_have(
        description: dict,
        have: list[str] | str | None = None,
        not_have: list[str] | str | None = None,
        be: list[str] | str | None = None,
        **kwargs,
    ) -> result:
        """Test column names match up with expected values"""
        have = list(have) if isinstance(have, str) else have
        not_have = list(not_have) if isinstance(not_have, str) else not_have
        be = list(be) if isinstance(be, str) else be
        checks: list[bool] = []
        present_columns = description["columns"].split("_^&^_")
        if have is not None:
            for col in have:
                checks.append(col in present_columns)
        if not_have is not None:
            for col in not_have:
                checks.append(col not in present_columns)
        if be is not None:
            checks.append(set(present_columns) == set(be))
            checks.append(len(present_columns) == len(be))
        return result(
            name="columns",
            success=all(checks),
            unexpected=present_columns if not all(checks) else None,
        )

    should_have_partial = partial(should_have, have=have, not_have=not_have, be=be)
    should_have_partial.required_metrics = {}
    return should_have_partial


def type_should(
    column: str,
    be: str | None = None,
    not_be: str | None = None,
    be_one_of: list[str] | None = None,
    **kwargs,
) -> Callable:
    """
    Test column type matches up with expected value.

    Note that this will expect *polars* style types, although does not require
    that they be case sensitive. For example, if testing a pandas dataframe for
    integer type, specify "int64" rather than, say, "int64[pyarrow]" or otherwise.
    """

    def should_be(
        description: dict,
        column: str,
        be: str | None,
        not_be: str | None,
        be_one_of: str | None,
        **kwargs,
    ) -> result:
        """
        Test column type matches up with expected value.

        Note that this will expect *polars* style types, although does not require
        that they be case sensitive. For example, if testing a pandas dataframe for
        integer type, specify "int64" rather than, say, "int64[pyarrow]" or otherwise.
        """
        checks: list[bool] = []
        col_type = description[f"type_{column}"]
        if be is not None:
            checks.append(be.lower() == col_type.lower())
        if not_be is not None:
            checks.append(be.lower() != col_type.lower())
        if be_one_of is not None:
            checks.append(col_type.lower() in [i.lower() for i in be_one_of])
        return result(
            name=f"type-of-{column}",
            success=all(checks),
            unexpected=col_type if not all(checks) else None,
        )

    should_be_partial = partial(
        should_be,
        column=column,
        be=be,
        not_be=not_be,
        be_one_of=be_one_of,
    )
    should_be_partial.required_metrics = {"type"}
    return should_be_partial


def average_difference_from_other_column_should(
    column: str,
    other_column: str,
    be_exactly: float | int | None = None,
    be_less_than: float | int | None = None,
    be_less_than_or_equal_to: float | int | None = None,
    be_greater_than: float | int | None = None,
    be_greater_than_or_equal_to: float | int | None = None,
    **kwargs,
) -> Callable:
    """
    Test that the average difference between column and other column are
    within designated bounds.
    """

    def should(
        description: dict,
        column: str,
        other_column: str,
        be_exactly: float | int | None = None,
        be_less_than: float | int | None = None,
        be_less_than_or_equal_to: float | int | None = None,
        be_greater_than: float | int | None = None,
        be_greater_than_or_equal_to: float | int | None = None,
        **kwargs,
    ) -> result:
        """
        Test that the average difference between column and other column are
        within designated bounds.
        """
        checks: list[bool] = []
        difference = description[f"mean_{column}"] = description[f"mean_{other_column}"]
        if be_exactly is not None:
            checks.append(difference == be_exactly)
        if be_less_than is not None:
            checks.append(difference < be_less_than)
        if be_less_than_or_equal_to is not None:
            checks.append(difference <= be_less_than_or_equal_to)
        if be_greater_than is not None:
            checks.append(difference > be_greater_than)
        if be_greater_than_or_equal_to is not None:
            checks.append(difference >= be_greater_than_or_equal_to)
        return result(
            name=f"average-difference-from-{column}-to-{other_column}",
            success=all(checks),
            unexpected=difference if not all(checks) else None,
        )

    should_partial = partial(
        should,
        column=column,
        other_column=other_column,
        be_exactly=be_exactly,
        be_less_than=be_less_than,
        be_less_than_or_equal_to=be_less_than_or_equal_to,
        be_greater_than=be_greater_than,
        be_greater_than_or_equal_to=be_greater_than_or_equal_to,
    )
    should_partial.required_metrics = {"mean"}
    return should_partial


def average_ratio_to_other_column_should(
    column: str,
    other_column: str,
    be_exactly: float | int | None = None,
    be_less_than: float | int | None = None,
    be_less_than_or_equal_to: float | int | None = None,
    be_greater_than: float | int | None = None,
    be_greater_than_or_equal_to: float | int | None = None,
    **kwargs,
) -> Callable:
    """
    Test that the average ratio between column and other column are
    within designated bounds (for instance, a value of 1 has a ratio
    of 0.1 to a value of 10)
    """

    def should(
        description: dict,
        column: str,
        other_column: str,
        be_exactly: float | int | None = None,
        be_less_than: float | int | None = None,
        be_less_than_or_equal_to: float | int | None = None,
        be_greater_than: float | int | None = None,
        be_greater_than_or_equal_to: float | int | None = None,
        **kwargs,
    ) -> result:
        """
        Test that the average ratio between column and other column are
        within designated bounds (for instance, a value of 1 has a ratio
        of 0.1 to a value of 10)
        """
        checks: list[bool] = []
        ratio = description[f"mean_{column}"] / description[f"mean_{other_column}"]
        if be_exactly is not None:
            checks.append(ratio == be_exactly)
        if be_less_than is not None:
            checks.append(ratio < be_less_than)
        if be_less_than_or_equal_to is not None:
            checks.append(ratio <= be_less_than_or_equal_to)
        if be_greater_than is not None:
            checks.append(ratio > be_greater_than)
        if be_greater_than_or_equal_to is not None:
            checks.append(ratio >= be_greater_than_or_equal_to)
        return result(
            name=f"average-ratio-between-{column}-and-{other_column}",
            success=all(checks),
            unexpected=ratio if not all(checks) else None,
        )

    should_partial = partial(
        should,
        column=column,
        other_column=other_column,
        be_exactly=be_exactly,
        be_less_than=be_less_than,
        be_less_than_or_equal_to=be_less_than_or_equal_to,
        be_greater_than=be_greater_than,
        be_greater_than_or_equal_to=be_greater_than_or_equal_to,
    )
    should_partial.required_metrics = {"mean"}
    return should_partial


possible_tests: dict[str, Callable] = {
    "mean_should": (mean_should := _range_check("mean")),
    "min_should": (min_should := _range_check("min")),
    "max_should": (max_should := _range_check("max")),
    "std_should": (std_should := _range_check("std")),
    "null_count_should": (null_count_should := _range_check("null_count")),
    "count_should": (count_should := _range_check("count")),
    "null_percentage_should": (
        null_percentage_should := _range_check("null_percentage")
    ),
    "columns_should": columns_should,
    "type_should": type_should,
    "row_count_should": row_count_should,
    "average_difference_from_other_column_should": average_difference_from_other_column_should,
    "average_ratio_to_other_column_should": average_ratio_to_other_column_should,
}



import json
from typing import Any, Callable

import fsspec

from wimsey.tests import possible_tests


def collect_tests(config: list[dict] | dict | list[Callable]) -> list[Callable]:
    """
    Take a configuration, and build out tests
    """
    list_config: list[dict] | list[Callable] = (
        config if isinstance(config, list) else [config]
    )
    if isinstance(list_config[0], Callable):
        return list_config
    tests: list[Callable] = []
    for item in list_config:
        test_name: str | None = item.get("test")
        test: Callable | None = None
        if test_name:
            test = possible_tests.get(item.get("test"))(**item)  # type: ignore[arg-type]
        if test is None:
            msg = (
                "Issue reading configuration, for at least one test, either no "
                "test is named, or a mispelt/unimplemented test is given"
            )
            raise ValueError(msg)
        tests.append(test)
    return tests


def read_config(path: str, storage_options: dict | None = None) -> list[Callable]:
    """
    Read a json or yaml configuration, and return list of test callables
    """
    storage_options_dict: dict = storage_options or {}
    config: dict
    with fsspec.open(path, "rt", **storage_options_dict) as file:
        contents = file.read()
    if path.endswith(".yaml"):
        try:
            import yaml

            config = parse_contents(yaml.safe_load(contents))
            return collect_tests(config)  # type: ignore[arg-type]
        except ImportError as exception:
            msg = (
                "It looks like you're trying to import a yaml configured "
                "test suite. This is supported but requires an additional "
                "install of pyyaml (`pip install pyyaml`)"
            )
            raise ImportError(msg) from exception
    config = parse_contents(json.loads(contents))
    return collect_tests(config)  # type: ignore[arg-type]


def parse_contents(contents: Any) -> list[dict] | dict:
    if isinstance(contents, list):
        return contents
    if isinstance(contents, dict):
        if isinstance(contents.get("tests"), list):
            return contents.get("tests")
        return contents
    msg = (
        "It looks like the json/yaml file parsed in is either invalid "
        "or doesn't match what's required for Wimsey to interpret tests. \n"
        "Hint: json/yaml file should either be a list of tests, a single test "
        "or a key/value pair with a 'tests' key relating to a list of tests"
    )
    raise ValueError(msg)



import polars as pl

from wimsey import execution
from wimsey import tests


def test_run_all_tests_produces_expected_result_object():
    tests_to_carry_out = [
        tests.max_should(column="a", be_less_than=10),
        tests.std_should(column="a", be_greated_than=0),
        tests.type_should(column="b", be_one_of=["string", "int64"]),
    ]
    df = pl.DataFrame({"a": [1, 2, 3], "b": ["hat", "bat", "cat"]})
    actual = execution.run_all_tests(df, tests_to_carry_out)
    assert actual.success is True
    for result in actual.results:
        assert result.success is True



import narwhals.stable.v1 as nw
from narwhals.stable.v1.typing import FrameT


@nw.narwhalify
def describe(
    df: FrameT,
    columns: list[str] | None = None,
    metrics: list[str] | None = None,
) -> dict[str, float]:
    """
    Outputs a dictionary for use in testing, mimicking polars 'describe' method.

    Note this code is adapted from polars own descrip function.
    """
    if not df.columns:
        return {}
    columns = columns or df.columns
    columns_to_check = [i for i in columns if i in df.columns]
    metrics = metrics or [
        "mean",
        "std",
        "min",
        "max",
        "type",
        "count",
        "null",
        "length",
    ]

    # Determine which columns should get std/mean/percentile statistics
    stat_cols = {c for c, dt in df.schema.items() if dt.is_numeric()}

    required_exprs: list = [
        nw.lit("_^&^_".join(df.columns)).alias("columns"),
    ]
    post_exprs: list = []
    if "mean" in metrics:
        required_exprs += [
            (nw.col(c).mean() if c in stat_cols else nw.lit(None)).alias(f"mean_{c}")
            for c in columns_to_check
        ]
    if "std" in metrics:
        required_exprs += [
            (nw.col(c).std() if c in stat_cols else nw.lit(None)).alias(f"std_{c}")
            for c in columns_to_check
        ]
    if "min" in metrics:
        required_exprs += [
            (nw.col(c).min() if c in stat_cols else nw.lit(None)).alias(f"min_{c}")
            for c in columns_to_check
        ]
    if "max" in metrics:
        required_exprs += [
            (nw.col(c).max() if c in stat_cols else nw.lit(None)).alias(f"max_{c}")
            for c in columns_to_check
        ]
    if "type" in metrics:
        required_exprs += [
            nw.lit(str(df.schema[c])).alias(f"type_{c}") for c in columns_to_check
        ]
    if "count" in metrics or "null" in metrics or "leghth" in metrics:
        required_exprs += [nw.col(*columns_to_check).count().name.prefix("count_")]
    if "null" in metrics or "length" in metrics:
        required_exprs += [
            nw.col(*columns_to_check).null_count().name.prefix("null_count_")
        ]
        post_exprs += [
            (
                nw.col(f"null_count_{c}")
                / (nw.col(f"count_{c}") + nw.col(f"null_count_{c}"))
            ).alias(f"null_percentage_{c}")
            for c in columns_to_check
        ]
        post_exprs += [
            (
                nw.col(f"count_{columns_to_check[0]}")
                + nw.col(f"null_count_{columns_to_check[0]}")
            ).alias("length")
        ]
    df_metrics = df.select(
        *required_exprs,
    ).with_columns(*post_exprs)
    try:
        return {k: v[0] for k, v in df_metrics.to_dict(as_series=False).items()}  # type: ignore[union-attr]
    except AttributeError:
        return {
            k: v[0]
            for k, v in df_metrics.collect().to_dict(as_series=False).items()  # type: ignore[union-attr]
        }



from typing import Callable, Any
from dataclasses import dataclass

from narwhals.typing import FrameT

from wimsey.dataframe import describe
from wimsey.tests import result
from wimsey.config import read_config, collect_tests


@dataclass
class final_result:
    success: bool
    results: list[result]


class DataValidationException(Exception): ...


def _as_set(val: Any) -> set:
    """
    Internal function, if val is none, return empty set,
    otherwise return set of just val
    """
    return {val} if val is not None else set()


def run_all_tests(df: FrameT, tests: list[Callable[[Any], result]]) -> final_result:
    columns: set[str] | None = set()
    metrics: set[str] | None = set()
    for test in tests:
        try:
            metrics |= test.required_metrics
            columns |= _as_set(test.keywords.get("column"))
            columns |= _as_set(test.keywords.get("other_column"))
        except AttributeError:
            columns = None  # fall back to calculating everything
            metrics = None
            break
    description: dict[str, Any] = describe(
        df,
        columns=list(columns),
        metrics=list(metrics),
    )
    results: list[result] = []
    for i_test in tests:
        results.append(i_test(description))
    return final_result(
        success=all(i.success for i in results),
        results=results,
    )


def test(
    df: FrameT, contract: str | list[dict] | dict, storage_options: dict | None = None
) -> final_result:
    """
    Carry out tests on dataframe and return results. This will *not* raise
    an exception on test failure, and will instead return a 'final_result'
    object, with a boolean 'success' field, and a detailed list of individual
    tests.

    If you want to halt processing in the event of a data contract failure,
    see `validate` function.
    """
    tests = (
        read_config(path=contract, storage_options=storage_options)
        if isinstance(contract, str)
        else collect_tests(contract)
    )
    return run_all_tests(df, tests)


def validate(
    df: FrameT, contract: str | list[dict] | dict, storage_options: dict | None = None
) -> FrameT:
    """
    Carry out tests on dataframe, returning original dataframe if tests are
    successful, and raising a DataValidationException in case of failure.
    """
    results = test(
        df=df,
        contract=contract,
        storage_options=storage_options,
    )
    if not results.success:
        failures: list[str] = [
            f"{i.name} (unexpected: {i.unexpected})"
            for i in results.results
            if not i.success
        ]
        newline = "\n - "
        msg = f"At least one test failed:\n - {newline.join(failures)}"
        raise DataValidationException(msg)
    return df
